# -*- coding: utf-8 -*-

# Aqua-Duct, a tool facilitating analysis of the flow of solvent molecules in molecular dynamic simulations
# Copyright (C) 2016-2017  Tomasz Magdziarz, Alicja Płuciennik, Michał Stolarczyk <info@aquaduct.pl>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# Works for PyMOL 1.4, 1.7, 1.8

ALPHA = 25.0
ALPHA_TRIANGLE = 17.0
BEGIN = 2.0
CHAR = 23.0
COLOR = 6.0
CONE = 27.0
CUSTOM_CYLINDER = 15.0
CYLINDER = 9.0
DEFAULT_ERROR = -1
DEFAULT_SUCCESS = None
DISABLE = 13.0
DOTWIDTH = 16.0
ELLIPSOID = 18.0
ENABLE = 12.0
END = 3.0
FONT = 19.0
FONT_AXES = 22.0
FONT_SCALE = 20.0
FONT_VERTEX = 21.0
LIGHTING = 2896.0
LINES = 1.0
LINEWIDTH = 10.0
LINE_LOOP = 2.0
LINE_STRIP = 3.0
NORMAL = 5.0
NULL = 1.0
POINTS = 0.0
QUADRIC = 26.0
SAUSAGE = 14.0
SPHERE = 7.0
STOP = 0.0
TRIANGLE = 8.0
TRIANGLES = 4.0
TRIANGLE_FAN = 6.0
TRIANGLE_STRIP = 5.0
VERTEX = 4.0
WIDTHSCALE = 11.0
